# -*- coding:utf-8 -*-


"""
This module contains definitions for different kinds of dialogs and related
components that are specific for this application.
"""


from PySide2 import QtWidgets
#
from . import INSTRUCTIONS, ABOUT
from ..dialogs import InfoDialog, FlexibleDialog


# #############################################################################
# ## HELP DIALOGS
# #############################################################################
class InstructionsDialog(InfoDialog):
    """
    Info dialog showing instructions
    """
    def __init__(self):
        """
        """
        super().__init__("INSTRUCTIONS", INSTRUCTIONS, print_msg=False)


class AboutDialog(InfoDialog):
    """
    Info dialog showing about section
    """
    def __init__(self):
        """
        """
        super().__init__("ABOUT", ABOUT, print_msg=False)


class KeymapsDialog(FlexibleDialog):
    """
    Info dialog showing keymap list
    """
    def __init__(self, mappings, parent=None):
        """
        """
        self.mappings = mappings
        super().__init__(parent=parent)

    def setup_ui_body(self, widget):
        """
        """
        lyt = QtWidgets.QVBoxLayout(widget)
        #
        self.list_widget = QtWidgets.QListWidget()
        for k, v in self.mappings.items():
            self.list_widget.addItem("{} ({})".format(k, v))
        lyt.addWidget(self.list_widget)


# #############################################################################
# ## SAVE DIALOGS
# #############################################################################
class SavedInfoDialog(InfoDialog):
    """
    Informative dialog telling about saved paths.
    """
    def __init__(self, save_dict, timeout_ms=500):
        """
        :param save_dict: A dict with ``mask_name: save_path`` pairs.
        """
        super().__init__("SAVED", self.save_dict_to_str(save_dict),
                         timeout_ms=timeout_ms,
                         header_style="font-weight: bold; color: green")

    @staticmethod
    def save_dict_to_str(save_dict):
        """
        """
        msg = "\n".join(["Saved {} to {}".format(k, v)
                         for k, v in save_dict.items()])
        return msg


class SaveWarningDialog(InfoDialog):
    """
    A dialog to be prompted when trying to delete unsaved changes.
    Usage example::

      self.dialog = SaveWarningDialog()
      user_wants_to_remove = bool(self.dialog.exec_())
      ...
    """

    def __init__(self):
        """
        """
        super().__init__("WARNING",
                         "DELETE unsaved changes?",
                         "YES, delete and continue", "NO, go back",
                         print_msg=False)
        self.reject_b.setDefault(True)


class SavedStateTracker:
    """
    Create one of these every time a new state is loaded, call ``edited`` when
    the state has been changed, and ``saved`` when saved.

    The ``saved`` function will optionally show an informative dialog.

    Then call ``delete`` when the state is intended to be deleted. The method
    makes sure that unsaved changes are only deleted with user's confirmation.

    Note for developers:

    State machines with callbacks are a classic recipe for spaghetti alla
    callback inferno. Here we didn't provide a structured way to handle GUI
    state, so we are applying this as high in the API as possible.
    It works NOW, but consider restructuring it if it gets in the way.
    """

    def __init__(self):
        """
        """
        self._dialog = None
        #
        self._has_unsaved_changes = False
        self._has_been_deleted = False

    def edit(self):
        """
        Call this any time the state that we want to track has been edited
        """
        # bug factory
        # assert not self._has_been_deleted, \
        #     "Deleted tracker cannot be further used! create a new one"
        self._has_unsaved_changes = True

    def save(self, saved_dict=None, ok_dialog_ms=1000):
        """
        Call this any time the state that we want to track has been saved
        """
        # bug factory
        # assert not self._has_been_deleted, \
        #     "Deleted tracker cannot be further used! create a new one"
        self._has_unsaved_changes = False
        if saved_dict is not None:
            self.dialog = SavedInfoDialog(saved_dict, ok_dialog_ms)
            self.dialog.show()

    def delete(self):
        """
        Call this when we intend to delete the information that we are
        tracking. If unsaved changes, it will prompt the user to continue.
        """
        # bug factory
        # assert not self._has_been_deleted, \
        #     "Deleted tracker cannot be further used! create a new one"
        if self._has_unsaved_changes:
            self.dialog = SaveWarningDialog()
            user_wants_delete = bool(self.dialog.exec_())
            if not user_wants_delete:
                # in this case the user hit this by mistake, so
                # we return False and do nothing
                return False
        # if we reach this point, it means that either we don't have
        # unsaved changes, or we have but the user is OK with deleting them
        self._has_been_deleted = True
        return True



# -*- coding:utf-8 -*-


"""
This module is a library of reusable, extendable widgets.
"""


import os
from PySide2 import QtCore, QtWidgets


# #############################################################################
# # BASIC WIDGETS
# #############################################################################
class FileList(QtWidgets.QWidget):
    """
    A file dialog button followed by a list that shows the files in the
    selected folder.
    """
    def __init__(self, label, parent=None,
                 default_path=None, extensions=None, sort=True):
        """
        :param extensions: A list of string terminations to match, or ``None``
          for match-all.
        :param default_path: If None, 'home' is picked as default.
        :param sort: If true, contents will always be shown sorted
        """
        super().__init__(parent)
        self.sort = sort
        self.dirpath = (os.path.expanduser("~") if default_path is None
                        else default_path)
        self.extensions = [""] if extensions is None else extensions
        self.label = label
        # create widgets
        self.file_button = QtWidgets.QPushButton(self.label)
        self.file_list = QtWidgets.QListWidget()
        # add widgets to layout
        self.main_layout = QtWidgets.QVBoxLayout(self)
        self.main_layout.addWidget(self.file_button)
        self.main_layout.addWidget(self.file_list)
        # connect
        self.file_button.pressed.connect(self._file_dialog_handler)
        # track filesystem changes
        self.file_watcher = QtCore.QFileSystemWatcher()
        self.file_watcher.fileChanged.connect(
            lambda: self.update_path(self.dirpath))
        self.file_watcher.directoryChanged.connect(
            lambda: self.update_path(self.dirpath))

    def update_path(self, dirname):
        """
        :param str dirname: The new directory path to be listed.
        """
        file_names = [f for f in os.listdir(dirname)
                      if f.lower().endswith(tuple(self.extensions))]
        self.file_list.clear()
        self.file_list.addItems(file_names)
        #
        self.file_watcher.removePath(self.dirpath)
        self.file_watcher.addPath(dirname)
        #
        self.dirpath = dirname
        if self.sort:
            self.file_list.sortItems(QtCore.Qt.AscendingOrder)

    def _file_dialog_handler(self):
        """
        Opens a file dialog which returns the selected path.
        """
        dirname = QtWidgets.QFileDialog.getExistingDirectory(
            self, self.label, self.dirpath)
        if dirname:
            self.update_path(dirname)


class CheckBoxGroup(QtWidgets.QWidget):
    """
    A group of ``CheckBox`` es
    """
    def __init__(self, parent=None, horizontal=False):
        """
        """
        super().__init__(parent=parent)
        ly = QtWidgets.QHBoxLayout() if horizontal else QtWidgets.QVBoxLayout()
        self.setLayout(ly)

    def add_box(self, name, tristate=False, initial_val=True):
        """
        :param bool tristate: If true, the added check box will have 3 states.
        """
        b = QtWidgets.QCheckBox(name)
        b.setTristate(tristate)
        b.setChecked(initial_val)
        self.layout().addWidget(b)

    def remove_box(self, idx):
        """
        :param int idx: Boxes are added in increasing index order, so this
          the lower this index the 'older' the box that is being removed.
        """
        b = self.layout().takeAt(idx)
        b.widget().setParent(QtWidgets.QWidget())  # this is needed...

    def state(self):
        """
        :returns: A list with all the current states, in index order.
        """
        ly = self.layout()
        return [ly.itemAt(i).widget().checkState() for i in range(ly.count())]


# #############################################################################
# # PAINTING SECTION
# #############################################################################
class RGBASpinbox(QtWidgets.QWidget):
    """
    A cluster of 4 [0-255] spin boxes, representing (and having) an RGBA
    color. Use ``self.connect`` to wire this widget to any method.
    """

    def __init__(self, initial_rgba=(0, 255, 0, 100), parent=None,
                 min_alpha=1, max_alpha=255):
        """
        """
        super().__init__(parent)
        self.main_layout = QtWidgets.QHBoxLayout(self)
        # create widgets
        self.label = QtWidgets.QLabel("RGBA:")
        r, g, b, a = initial_rgba
        self.r = self._make_box(initial=r)
        self.g = self._make_box(initial=g)
        self.b = self._make_box(initial=b)
        self.a = self._make_box(a, min_alpha, max_alpha)
        # add widgets to layout
        self.main_layout.addWidget(self.label)
        self.main_layout.addWidget(self.r)
        self.main_layout.addWidget(self.g)
        self.main_layout.addWidget(self.b)
        self.main_layout.addWidget(self.a)
        # connect widgets to color listener
        self.r.valueChanged.connect(
            lambda: self._set_currentcolor_stylesheet())
        self.g.valueChanged.connect(
            lambda: self._set_currentcolor_stylesheet())
        self.b.valueChanged.connect(
            lambda: self._set_currentcolor_stylesheet())
        self.a.valueChanged.connect(
            lambda: self._set_currentcolor_stylesheet())
        #
        self._set_currentcolor_stylesheet()

    @staticmethod
    def _make_box(initial=100, minimum=0, maximum=255, step=8):
        """
        """
        b = QtWidgets.QSpinBox()
        b.setRange(minimum, maximum)
        b.setSingleStep(step)
        b.setValue(initial)
        return b

    def get_current_rgba(self):
        """
        :returns: Current state as ``(r, g, b, a)`` numeric tuple.
        """
        r = self.r.value()
        g = self.g.value()
        b = self.b.value()
        a = self.a.value()
        return(r, g, b, a)

    def _set_currentcolor_stylesheet(self):
        """
        """
        r, g, b, _ = self.get_current_rgba()
        ssheet = "QSpinBox {color: rgb(%d, %d, %d)}" % (r, g, b)
        self.r.setStyleSheet(ssheet)
        self.g.setStyleSheet(ssheet)
        self.b.setStyleSheet(ssheet)
        self.a.setStyleSheet(ssheet)

    def connect(self, fn):
        """
        :param fn: A function to connect this widget to. It must have the
          following signature``fn(idx, r, g, b, a)``.

        When calling ``self.connect(f)``, any value changes in R, G, B or
        A will trigger ``f(r, g, b, a)`` with the changed values
        """
        # NOTE THAT LAMBDAS WITH SELF INSIDE WILL PREVENT THE WIDGET
        # TO BE GARBAGE-COLLECTED: https://stackoverflow.com/a/48501804/4511978
        # THIS IS BAD FOR WIDGETS THAT GET INSTANTIATED MANY TIMES
        self.r.valueChanged.connect(lambda: fn(*self.get_current_rgba()))
        self.g.valueChanged.connect(lambda: fn(*self.get_current_rgba()))
        self.b.valueChanged.connect(lambda: fn(*self.get_current_rgba()))
        self.a.valueChanged.connect(lambda: fn(*self.get_current_rgba()))


class MaskPaintForm(QtWidgets.QWidget):
    """
    This widget contains one section for the masks and one for the painter.
    The mask section contains a set of elements, one per mask. A radio button
    selects the currently active mask, and each mask features an RGBA box
    and a thershold slider. The painter section contains a ComboBox to select
    the painter type, and a slider for the painter size.

    To use it in specific applications override ``button_pressed,
    combo_box_changed, rgba_box_changed...``
    """
    def __init__(self, brush_names, max_brush_size=100, parent=None,
                 thresh_min=0, thresh_max=1, thresh_num_steps=100,
                 min_alpha=1, max_alpha=255):
        """
        """
        super().__init__(parent)
        # self.masks = masks
        self.brush_names = brush_names
        self.max_brush_size = max_brush_size
        self.thresh_min = thresh_min
        self.thresh_max = thresh_max
        self.thresh_num_steps = thresh_num_steps
        self.min_alpha = min_alpha
        self.max_alpha = max_alpha
        #
        self._buttons = []
        self._boxes = []
        self._labels = []
        self._sliders = []
        #
        self.paint_button_group = QtWidgets.QButtonGroup(self)
        self._define_layout()
        # add connections
        # NOTE THAT LAMBDAS WITH SELF INSIDE WILL PREVENT THE WIDGET
        # TO BE GARBAGE-COLLECTED: https://stackoverflow.com/a/48501804/4511978
        # THIS IS BAD FOR WIDGETS THAT GET INSTANTIATED MANY TIMES
        self.paint_button_group.buttonClicked.connect(
            lambda _: self.button_pressed(
                self.paint_button_group.checkedButton()))
        self.brush_combo_box.currentIndexChanged.connect(
            self.brush_type_changed)
        self.brush_size_slider.valueChanged.connect(self._update_brush_size)

    def _define_layout(self):
        """
        """
        # make brush selector and size widgets
        self.brush_combo_box = QtWidgets.QComboBox()
        self.brush_combo_box.addItems(self.brush_names)
        self.brush_size_label = QtWidgets.QLabel()
        self.brush_size_slider = QtWidgets.QSlider(None)  # vertical
        self.brush_size_slider.setMinimum(1)
        self.brush_size_slider.setMaximum(self.max_brush_size)
        self.brush_size_slider.setValue(self.max_brush_size // 2)
        self.brush_size_slider.setSingleStep(1)
        self._set_brush_size_label(self.brush_size_slider.value())
        # Layout hierarchy:
        main_layout = QtWidgets.QHBoxLayout()
        # this is modified with add/remove
        self.masks_layout = QtWidgets.QVBoxLayout()
        # make sub-layout with [brush_type; brush_size]
        brush_layout = QtWidgets.QVBoxLayout()
        brush_layout.addWidget(self.brush_combo_box)
        brush_layout.addWidget(self.brush_size_label)
        brush_layout.addWidget(self.brush_size_slider)
        brush_layout.setAlignment(QtCore.Qt.AlignRight)
        # connect layout hierarchy
        main_layout.addLayout(self.masks_layout)
        main_layout.addLayout(brush_layout)
        self.setLayout(main_layout)

    def add_item(self, name, rgba, slider_visible=True, activate=False):
        """
        Add an element to the 'mask' section, with the given name and color.

        :param slider_visible: If false, the slider will be still there but
          hidden.
        :param activate: Once created, select this item in the radio buttons.
        """
        # sub-layout with 2-row elts: [button, colordialog; label, threshold]
        but = QtWidgets.QRadioButton(name)
        box = RGBASpinbox(rgba, None, self.min_alpha, self.max_alpha)
        sl = QtWidgets.QSlider(None, orientation=QtCore.Qt.Horizontal)
        lbl = QtWidgets.QLabel()
        sl.setVisible(slider_visible)
        lbl.setVisible(slider_visible)
        sl.setMinimum(0)
        sl.setMaximum(self.thresh_num_steps)
        sl.setSingleStep(1)
        sl.setValue(sl.maximum() // 2)
        #
        self._buttons.append(but)
        self._boxes.append(box)
        self._labels.append(lbl)
        self._sliders.append(sl)
        # local layout hierarchy
        lyt = QtWidgets.QVBoxLayout()
        top = QtWidgets.QHBoxLayout()
        bottom = QtWidgets.QHBoxLayout()
        top.addWidget(but)
        top.addWidget(box)
        bottom.addWidget(lbl)
        bottom.addWidget(sl)
        lyt.addLayout(top)
        lyt.addLayout(bottom)
        # add local hierarchy to main layout and button to group
        self.masks_layout.addLayout(lyt)
        self.paint_button_group.addButton(but)
        # add connections
        box.connect(
            lambda r, g, b, a: self._handle_rgba_box_changed(box, r, g, b, a))
        sl.valueChanged.connect(
            lambda val: self._handle_threshold_slider_changed(sl, val))
        # initialize label
        self._set_thresh_label(lbl, sl.value())
        #
        if activate:
            but.click()

    def remove_item(self, idx):
        """
        Remove an element from the 'mask' section by index. Indexes are in
        increasing order, so lowest is oldest.
        """
        # remove from self placeholders
        but = self._buttons.pop(idx)
        box = self._boxes.pop(idx)
        lbl = self._labels.pop(idx)
        sl = self._sliders.pop(idx)
        lyt = self.masks_layout.takeAt(idx)
        # reassign dummy parent to remove from QT placeholders
        w = QtWidgets.QWidget()  # this is needed...
        but.setParent(w)
        box.setParent(w)
        lbl.setParent(w)
        sl.setParent(w)
        lyt.setParent(w)

    def slider_to_p_val(self, sl_val):
        """
        Since the slider goes from 0 to ``thresh_num_steps``, this function
        linearly interpolates the, so that 0 maps to ``thresh_min`` and
        ``thresh_num_steps`` maps to ``thresh_max``. Note that min does not
        neccesarily have to be smaller than max.

        :param int sl_val: The actual slider value from 0 to num_steps.
        :returns: The converted and interpolated value.
        """
        delta = float(sl_val) / self.thresh_num_steps
        pval = self.thresh_min + delta * (self.thresh_max - self.thresh_min)
        return pval

    def _set_thresh_label(self, lbl, sl_val):
        """
        """
        t = self.slider_to_p_val(sl_val)
        lbl.setText("Thresh: {:.10f}".format(t))
        return t

    def _set_brush_size_label(self, sl_val):
        """
        """
        self.brush_size_label.setText("Brush size: {}".format(sl_val))

    def _handle_threshold_slider_changed(self, sl, sl_val):
        """
        """
        idx = self._sliders.index(sl)
        t = self._set_thresh_label(self._labels[idx], sl_val)
        self.threshold_slider_changed(idx, t)

    def _handle_rgba_box_changed(self, box, r, g, b, a):
        """
        """
        idx = self._boxes.index(box)
        self.rgba_box_changed(idx, r, g, b, a)

    def _update_brush_size(self, sl_val):
        """
        """
        self._set_brush_size_label(sl_val)
        self.brush_size_changed(sl_val)

    def button_pressed(self, but):
        """
        Override me!

        :param int idx: Starts with 0 and respects ordering given at
          construction. So when overriding this method, you can assume
          that 0 will correspond to the firstly added element, and
          so on. Implementation example::

            i = self._buttons.index(but)
            print("button pressed: >>>", i, but.text())
        """
        pass

    def brush_type_changed(self, idx):
        """
        Override me!

        :param int idx: Starts with 0 and respects ordering given at
          construction. So when overriding this method, you can assume
          that 0 will correspond to the firstly added element, and
          so on.
        """
        pass

    def brush_size_changed(self, sz):
        """
        Override me!
        """
        pass

    def rgba_box_changed(self, idx, r, g, b, a):
        """
        Override me!
        """
        pass

    def threshold_slider_changed(self, idx, val):
        """
        Override me!
        """
        pass


# #############################################################################
# # SAVING SECTON
# #############################################################################
class SaveForm(QtWidgets.QWidget):
    """
    A formulary providing functionality for selecting what to save, where to
    save, the output suffix and overwriting policy.
    """

    SAVE_TEXT = "Save\nselected"
    DIALOG_TEXT = "Output\nfolder"
    OVERWRITE_TEXT = "Overwrite\nsaved"

    def __init__(self, parent=None, default_path=None):
        """
        :param str default_path: If not given, 'home' is picked as default.
        """
        super().__init__(parent)
        self.save_path = (os.path.expanduser("~") if default_path is None
                          else default_path)
        # create widgets
        self.save_group = CheckBoxGroup()
        self.text_boxes = QtWidgets.QVBoxLayout()
        self.file_dialog_button = QtWidgets.QPushButton(self.DIALOG_TEXT)
        self.save_button = QtWidgets.QPushButton(self.SAVE_TEXT)
        self.overwrite_button = QtWidgets.QCheckBox(self.OVERWRITE_TEXT)
        self.overwrite_button.setChecked(True)
        # build layout hierarchy
        self.main_layout = QtWidgets.QHBoxLayout()
        self.main_layout.addWidget(self.save_group)
        self.main_layout.addLayout(self.text_boxes)
        #
        lyt = QtWidgets.QVBoxLayout()
        lyt.addWidget(self.file_dialog_button)
        lyt.addWidget(self.save_button)
        lyt.addWidget(self.overwrite_button)
        self.main_layout.addLayout(lyt)
        self.setLayout(self.main_layout)
        # add connections
        self.file_dialog_button.clicked.connect(self._change_save_path)
        self.save_button.clicked.connect(self._handle_save_masks)

    def add_checkbox(self, checkbox_name, initial_val=True, initial_txt=None):
        """
        Adds an element that can be selected to be saved.

        :param checkbox_name: The element identifier
        :param initial_txt: The initial suffix to be appended to the files. If
          none is given, the ``checkbox_name`` is picked as default. The user
          can change this from the GUI.
        """
        if initial_txt is None:
            initial_txt = checkbox_name
        tb = QtWidgets.QLineEdit(initial_txt)
        #
        self.save_group.add_box(checkbox_name, False, initial_val)
        self.text_boxes.addWidget(tb)

    def _change_save_path(self):
        """
        """
        dir_name = QtWidgets.QFileDialog.getExistingDirectory(
            self, self.DIALOG_TEXT, self.save_path)
        if dir_name:
            self.save_path = dir_name

    def _handle_save_masks(self):
        """
        """
        states = [s == QtCore.Qt.CheckState.Checked
                  for s in self.save_group.state()]
        suffixes = [self.text_boxes.itemAt(i).widget().text()
                    for i in range(self.text_boxes.count())]
        overwrite = self.overwrite_button.isChecked()
        self.save_masks(states, suffixes, overwrite)

    def save_masks(self, states, suffixes, overwrite):
        """
        :param states: A list with booleans, representing the checkbox
          states for the contained elements.
        :param suffixes: A list with the corresponding suffixes
        :param overwrite: A boolean determining whether the 'overwrite'
          checkbox has been activated.

        Override me!
        """
        print("check boxes:", states, "suffixes:", suffixes,
              "out path:", self.save_path, "overwrite:", overwrite)



# -*- coding:utf-8 -*-


"""
This module contains helper functions and utilities that may be used anywhere
else in the project.
"""


import os
import itertools
from pathlib import Path
#
import numpy as np
import randomcolor
from PIL import Image
import exifread
from PySide2 import QtGui


# #############################################################################
# ##
# #############################################################################
class RandomColorGenerator(randomcolor.RandomColor):
    """
    Flexible generator for nice random colors. For more details check
    https://pypi.org/project/randomcolor/

    Usage example::
      r, g, b = next(RandomColorGenerator().generate(form="rgbArray"))
    """

    def generate(self, hue=None, luminosity=None, count=1, form="rgbArray"):
        """
        :param form: Popular ones: ``rgbArray, rgba, hex, rgb``
        :returns: A generator with ``count`` random colors.

        Overriden to return a generator instead of a list. Source:
          https://github.com/kevinwuhoo/randomcolor-py
        """
        for _ in range(count):
            # First we pick a hue (H)
            H = self.pick_hue(hue)
            # Then use H to determine saturation (S)
            S = self.pick_saturation(H, hue, luminosity)
            # Then use S and H to determine brightness (B).
            B = self.pick_brightness(H, S, luminosity)
            # Then we return the HSB color in the desired format
            yield self.set_format([H, S, B], form)


def unique_filename(path, suffix="_({})", max_iters=10000):
    """
    Given a path, returns the same path if unique, or adds ``(N)`` before the
    extension to make it unique, for ``N`` being the lowest integer possible
    starting from 1.
    """
    if not Path(path).is_file():
        return path
    else:
        prefix, ext = os.path.splitext(path)
        for i in itertools.count(start=1, step=1):
            p = prefix + suffix.format(i) + ext
            if not Path(p).is_file():
                return p
            assert i < max_iters, "max no. of iters reached!"


# #############################################################################
# ## NUMPY <-> QT_PIXMAP INTERFACING
# #############################################################################
def rgb_arr_to_rgb_pixmap(arr):
    """
    :param arr: Expects a ``np.uint8(h, w, 3)`` array.
    :returns: A ``QtGui.QPixmap`` in format ``RGB888(w, h)``.
    """
    h, w, c = arr.shape
    assert c == 3, "Only np.uint8 arrays of shape (h, w, 3) expected!"
    img = QtGui.QImage(arr.data, w, h,
                       arr.strides[0], QtGui.QImage.Format_RGB888)
    pm = QtGui.QPixmap.fromImage(img)
    return pm


def bool_arr_to_rgba_pixmap(arr, rgba=(255, 0, 0, 255)):
    """
    :param arr: Expects a ``np.bool(h, w)`` array.
    :param rgba: 4 values between 0 and 255. Alpha=255 means full opacity.
    :returns: A ``QtGui.QPixmap`` in format ``RGBA8888(w, h)``, where the
      ``false`` values are all zeros and the ``true`` values have the specified
      ``rgba`` color.
    """
    # When painting ``(r, g, b, 0)`` Qt actually paints ``(0, 0, 0, 0)``. The
    # workaround of inverting all pixel values before and after painting
    # didn't work... TODO
    # forum.qt.io/topic/73787/qimage-qpixmal-loses-alpha-color-when-drawing/5
    # also topic/88000/qpainter-loosing-color-of-transparent-pixels-critical
    assert rgba[-1] > 0, "Alpha can't be zero, Qt will delete all :("
    h, w = arr.shape
    marr = np.zeros((h, w, 4), dtype=np.uint8)
    y_idxs, x_idxs = np.where(arr)
    marr[y_idxs, x_idxs] = rgba
    # HERE WOULD COME THE BUGFIX: try harder the "invert colors" approach?
    #
    img = QtGui.QImage(marr.data, w, h, marr.strides[0],
                       QtGui.QImage.Format_RGBA8888)
    #
    pm = QtGui.QPixmap.fromImage(img)
    return pm


def pixmap_to_arr(pm, img_format=QtGui.QImage.Format_RGBA8888):
    """
    :param pm: A Pixmap to be converted
    :param img_format: The ``QtGui.QImage`` format that ``pm`` corresponds to.
      https://doc.qt.io/qtforpython/PySide2/QtGui/QImage.html#image-formats
    :returns: A ``np.uint8(h, w, C)`` array, where the number of channels ``C``
      depends on the image format.

    ..note::
      Pixmaps are in format (w, h, ...) but arrays are returned
      in ``(h, w, ...)``, as usual for numpy
    """
    img = pm.toImage().convertToFormat(img_format)
    w, h = img.size().toTuple()
    arr = np.array(img.constBits()).reshape(h, w, -1)
    return arr


# #############################################################################
# ## IMAGE I/O
# #############################################################################
def load_exif(img_path):
    """
    :returns: A dictionary with the EXIF data contained at ``img_path``.
    """
    with open(img_path, "rb") as f:
        exif_dict = exifread.process_file(f)
        return exif_dict


def load_img_and_exif(img_path: str, as_np_array=True,
                      ignore_alpha=True):
    """
    Loads the image at given path using PIL, and its EXIF data. If the EXIF
    data contains extra info about orientation, also rotates the image
    accordingly.

    :param as_np_array: If true, the image will be converted from PIL format
      to np via ``np.asarray(image)``
    :param ignore_alpha: If the type of the image is ``RGBA``, it will be
      converted to ``RGB``.

    :returns: A tuple ``(image, exif_dict)``.

    Inspired in https://stackoverflow.com/a/26928142
    """
    exif_dict = load_exif(img_path)
    image = Image.open(img_path)
    if ignore_alpha and image.mode == "RGBA":
        image = image.convert("RGB")

    # Img orientation seems to have tag 274
    try:
        orientation: int = exif_dict["Image Orientation"].values[0]
        if orientation == 3:
            image = image.rotate(180, expand=True)
        elif orientation == 6:
            image = image.rotate(270, expand=True)
        elif orientation == 8:
            image = image.rotate(90, expand=True)
    except KeyError:
        pass
    if as_np_array:
        image = np.asarray(image)
    return image, exif_dict



# -*- coding:utf-8 -*-


"""
Object composite commands are intended for composite objects (like e.g.
a train of points). The main difference with the regular commands is that
they implement a ``state`` method (which e.g. for a train of points returns
a list of (x, y) tuples), and a ``clear`` method, which allows to 'remove'
the object without having to roll back or break the undo queue.
"""


from PySide2 import QtCore, QtGui
from .commands import CompositeCommand, UndoableLambda


# #############################################################################
# ## HELPERS
# #############################################################################


# #############################################################################
# ## OBJECT COMPOSITE COMMANDS
# #############################################################################
class PointList(CompositeCommand):
    """
    This class provides functionality to add circles to a scene (optionally
    connected by lines), and to return their centers as a list of ``(x, y)``
    positions. It also
    """
    COMMAND_NAME = "Draw point list"

    def __init__(self, scene, diameter, fill_rgba=(100, 100, 100, 100),
                 contour_rgba=(0, 0, 0, 255),
                 draw_lines_between_dots=False,
                 comp_mode=QtGui.QPainter.CompositionMode_SourceOver,
                 parent=None):
        """
        :scene: A pointer to the ``QGraphicsScene`` to add points to.
        :fill_rgba: inner circle (and optionally line) color.
        :contour_rgba: circle contour color
        :draw_lines_between_dots: If true, connect the dots
        :comp_mode: ``SourceOver`` means alphas get added. Check Qt composition
          modes.
        """
        super().__init__(parent)
        #
        self.scene = scene
        self.diameter = diameter
        self.comp_mode = comp_mode
        #
        self.fill_rgba = fill_rgba
        self.contour_rgba = contour_rgba
        fill_color = QtGui.QColor(*fill_rgba)
        contour_color = QtGui.QColor(*contour_rgba)
        self.dot_brush = QtGui.QBrush(fill_color, bs=QtCore.Qt.SolidPattern)
        self.dot_pen = QtGui.QPen(contour_color)
        # "all points" can only grow, it will remember even if we undo
        # "points" will show the "active", not deleted points only
        self._all_points = []
        self.points = []
        #
        self.with_lines = draw_lines_between_dots
        r, g, b, a = fill_rgba
        line_color = QtGui.QColor(r, g, b, a // 2)
        self.line_pen = QtGui.QPen(line_color)
        self._lines = {}

    def state(self):
        """
        :returns: A list in the form ``[(x1, y1), ...]`` with the center
          of the currently active points.
        """
        centers = [elt.rect().center().toTuple() for elt in self.points]
        return centers

    def clear(self):
        """
        Removes all the active points from the datastructure and the scene.
        """
        while self.points:
            pmi = self.points.pop()
            self.scene.removeItem(pmi)
            try:
                line = self._lines[pmi]
                self.scene.removeItem(line)
            except KeyError:
                pass

    def action(self, x, y, undo_stack=None):
        """
        Add a new point at given position.
        :param undo_stack: If given, this action is added to the undo stack.
        """
        d = self.diameter
        radius = float(d) / 2
        rect = QtCore.QRect(*(x - radius, y - radius, d, d))
        pmi = self.scene.addEllipse(rect, pen=self.dot_pen,
                                    brush=self.dot_brush)
        #
        self._all_points.append(pmi)
        self.points.append(pmi)
        #
        if self.with_lines:
            centers = self.state()
            if len(centers) >= 2:
                x1, y1 = centers[-2]
                x2, y2 = centers[-1]
                line_pmi = self.scene.addLine(x1, y1, x2, y2,
                                              pen=self.line_pen)
                self._lines[pmi] = line_pmi
        #
        if undo_stack is not None:
            # deep copy is needed
            pmis_before = list(self.points[:-1])
            pmis_after = list(self.points)
            #
            cmd = UndoableLambda(
                "Draw point", lambda: self._set_active_pmis(pmis_before),
                lambda: self._set_active_pmis(pmis_after))
            undo_stack.push(cmd)

    def _set_active_pmis(self, pmis):
        """
        Given a collection of ``pmi`` s that exist in ``self._all_points``,
        set those, and only those, as active.
        """
        assert all([p in self._all_points for p in pmis]), \
            "All pmis must preexist!"
        self.clear()
        for pmi in pmis:
            self.scene.addItem(pmi)
            self.points.append(pmi)
            try:
                line_pmi = self._lines[pmi]
                self.scene.addItem(line_pmi)
            except KeyError:
                pass

    def redo(self):
        """
        Unused
        """
        self._set_active_pmis(self._all_points)

    def undo(self):
        """
        Unused
        """
        self._set_active_pmis([])

    def finish(self, undo_stack=NotImplemented):
        """
        Simply sets ``self.finished`` to true, because the undo stack gets
        the separate ``actions`` instead of the whole composite one.
        """
        self.finished = True


# #############################################################################
# ## OBJECT CONTAINER
# #############################################################################
class ObjectContainer:
    """
    This class is a Mixin. When a ``QGraphicsScene`` inherits from it, it
    acquires functionality to add multiple composite objects from this module.
    """
    def __init__(self):
        """
        """
        # objects come on the top of masks
        self.objects = {}
        self._current_object_action = None

    def close_current_object_action(self, undo_stack=None):
        """
        If there is an open object action, closes it and optionally adds it to
        the undo stack
        """
        cmd = self._current_object_action
        if cmd is not None:
            cmd.finish(undo_stack)
            self._current_object_action = None

    def object_action(self, obj_class, action_args, obj_instantiation_args,
                      undo_stack=None):
        """
        This function implements a protocol to add composite objects to the
        scene.

        1. If a different composite action was running, it closes it and starts
            this one, adding it to ``self.objects``.
        2. If no composite action was running, starts this one and adds it.
        3. If ``obj_class`` is already running, does nothing here.

        In all cases, including if ``obj_class`` was already running,
        performs the action ``obj.action(*action_args)``.

        .. note::
          The scene simply calls the object's action. The object is responsible
          for keeping track of the ``scene items`` it generates, and also
          removing/adding them to the scene when needed.

        :param obj_instantiation_args: If this action needs to be started, it
          will be called via ``cmd = action_class(*instantiation_args)``
        :param action_args: The action will be called with this args.

        Usage example::

          # adds a point to the existing cloud or starts one otherwise
          scene.object_action(ExamplePointCloud, [x, y],
                              [cloud_color, points_size...])
          # check the state of the last added point cloud (this one):
          scene.objects[ExamplePointCloud][-1].state()
        """
        cmd = self._current_object_action
        # if changed to this action without releasing the prior one, release it
        action_changed = obj_class is not cmd.__class__
        cmd_finished = cmd is not None and cmd.finished
        #
        if action_changed:
            self.close_current_object_action(undo_stack)
            cmd = self._current_object_action  # this should be None
        # if no open action of this class, create
        if cmd is None or cmd_finished:
            cmd = obj_class(*obj_instantiation_args)
            self._current_object_action = cmd
            self.objects.setdefault(obj_class, []).append(cmd)
        cmd.action(*action_args)



# -*- coding:utf-8 -*-


"""
This module contains the ``QGraphicsScene+QGraphicsView`` binomial, typically
used in Qt apps to display and navigate images, together with some specific
functionality to annotate.
"""


import numpy as np
from PySide2 import QtCore, QtWidgets, QtGui
#
from .utils import RandomColorGenerator, rgb_arr_to_rgb_pixmap, \
    bool_arr_to_rgba_pixmap, pixmap_to_arr
from .mouse_event_manager import MouseEventManager
from .objects import ObjectContainer


# #############################################################################
# ## SCENE (PAINT ETC)
# #############################################################################
class MaskedImageScene(QtWidgets.QGraphicsScene, ObjectContainer):
    """
    Basic area that allows to display a color image, together with a set of
    binary masks on top of it.
    """
    DEFAULT_MASK_ALPHA = 100  # 255 is opaque. Used if no colors are specified

    def __init__(self, img_arr=None, parent=None):
        """
        :param img_arr: See ``update_image``
        """
        # super().__init__(parent)
        QtWidgets.QGraphicsScene.__init__(self, parent)
        ObjectContainer.__init__(self)
        #
        self.img_pmi = None
        self.h = None
        self.w = None
        #
        self.mask_pmis = {}  # pmi_ref : (r, g, b, a)
        if img_arr is not None:
            self.update_image(img_arr)

    def update_image(self, img_arr):
        """
        Clears whole scene, and adds the given numpy array as Pixmap.

        :param img_arr: A ``np.uint8(h, w [, ?])`` array.
        """
        pm = rgb_arr_to_rgb_pixmap(img_arr)
        self.clear()
        self.img_pmi = self.addPixmap(pm)
        #
        self.h, self.w = img_arr.shape[:2]
        self.setSceneRect(0, 0, self.w, self.h)

    def num_items(self):
        """
        Number of items in this scene, ordered from foreground to background.
        """
        return len(self.items(QtCore.Qt.DescendingOrder))

    def items(self, ascending=True):
        """
        :returns: This scene's items.
        :param ascending: If true, items are given from background to
          foreground.
        """
        o = QtCore.Qt.AscendingOrder if ascending else QtCore.Qt.AscendingOrder
        return super().items(o)

    def add_mask(self, mask_arr, rgba=None, item_on_top=None):
        """
        :param mask_arr: A ``np.bool(h, w)`` array.
        :param item_on_top: If given, mask will be added underneath that item.
          Otherwise will be added on top of item stack.
        :returns: The added ``PixmapItem``.
        """
        # sanity check mask
        assert mask_arr.dtype == np.bool, "Mask must be np.bool(h, w)!"
        assert len(mask_arr.shape) == 2, "Mask must be np.bool(h, w)!"
        assert mask_arr.shape == (self.h, self.w), \
            "Mask must have same (h, w) as image!"
        # sanity check color
        if rgba is None:
            r, g, b = next(RandomColorGenerator().generate(form="rgbArray"))
            rgba = (r, g, b, self.DEFAULT_MASK_ALPHA)
        assert all([0 <= c <= 255 for c in rgba]), \
            "RGBA must be in [0, 255] range!"
        # add pixmap: if this fails, the method raises with no side effect
        pm = bool_arr_to_rgba_pixmap(mask_arr, rgba)
        pmi = self.addPixmap(pm)
        self.mask_pmis[pmi] = rgba
        # now we have side FX: if this fails roll back the add Pixmap
        if item_on_top is not None:
            try:
                pmi.stackBefore(item_on_top)
            except Exception:
                # roll back addition
                del self.mask_pmis[pmi]
                self.removeItem(pmi)
                raise RuntimeError(
                    "Invalid other_item? {}".format(item_on_top))
        #
        return pmi

    def remove_mask(self, pmi):
        """
        :param pmi: The PixmapItem to remove. It has to be a mask added via
          ``add_mask``
        """
        # check that exists in our dict
        rgba = self.mask_pmis[pmi]
        #
        self.removeItem(pmi)
        del self.mask_pmis[pmi]
        return rgba

    def replace_mask_pmi(self, pmi, new_mask_arr, new_rgba=None):
        """
        If we call ``remove_mask`` and then ``add_mask`` fails, we will lose
        the removed mask forever. This method updates the mask in an atomary
        way: either succeeds or does nothing.

        .. warning::
          The input ``pmi`` gets removed from the scene and the reference
          becomes invalid. Use the reference returned by this function instead.
        """
        # the problem is if we successfully remove and then fail to add. A
        # fix is to first add below the to-be-removed, and then remove
        old_rgba = self.mask_pmis[pmi]
        if new_rgba is None:
            new_rgba = old_rgba
        new_pmi = self.add_mask(new_mask_arr, rgba=new_rgba, item_on_top=pmi)
        self.remove_mask(pmi)
        return new_pmi

    def mask_as_bool_arr(self, pmi):
        """
        Asserts that the given ``pmi`` is in ``self.mask_pmis``, and returns
        the map as ``np.bool(h, w)`` array, in which all non-zero values are
        true.
        """
        assert pmi in self.mask_pmis, "Given Item is not in mask_pmis!"
        # a has shape (h, w, 4) where 4->RGBA
        pixmap_format = QtGui.QImage.Format_RGBA8888
        arr = pixmap_to_arr(pmi.pixmap(), pixmap_format)
        mask = (arr > 0).any(axis=-1)
        return mask


# #############################################################################
# ## VIEW (EVENTS ETC)
# #############################################################################
class DisplayView(MouseEventManager, QtWidgets.QGraphicsView):
    """
    In Qt applications, it is usual to wrap a scene with a view. This allows
    to dynamically and easily change the perspective on the scene.
    """
    def __init__(self, scene=None, parent=None, scale_percent=15):
        """
        :param scene: The scene to view. It can be set also afterwards via
          ``self.setScene``.
        """
        QtWidgets.QGraphicsView.__init__(self, scene, parent)
        MouseEventManager.__init__(self, True)
        self.scale_factor_up = 1 + float(scale_percent) / 100
        self.scale_factor_down = 1.0 / self.scale_factor_up
        #
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        # Set Anchors
        self.setTransformationAnchor(QtWidgets.QGraphicsView.NoAnchor)
        self.setResizeAnchor(QtWidgets.QGraphicsView.NoAnchor)
        #
        if scene is not None:
            self.fit_in_scene()

    def fit_in_scene(self):
        """
        Moves perspective so that the whole scene can be seen in view.
        """
        self.fitInView(self.scene().sceneRect(), QtCore.Qt.KeepAspectRatio)

    def zoom(self, pos_x, pos_y, zoom_out=False):
        """
        Source for wheel zoom: https://stackoverflow.com/a/29026916/4511978
        """
        # Save the scene pos
        qpos = QtCore.QPoint(pos_x, pos_y)
        old_pos = self.mapToScene(qpos)
        # Zoom
        scale_factor = (self.scale_factor_down if zoom_out
                        else self.scale_factor_up)
        self.scale(scale_factor, scale_factor)
        # Get the new position
        new_pos = self.mapToScene(qpos)
        # Move scene to old position
        delta = new_pos - old_pos
        self.translate(delta.x(), delta.y())

    def shift_view(self, delta_x, delta_y):
        """
        Move perspective to shift through the scene
        """
        self.horizontalScrollBar().setValue(
            self.horizontalScrollBar().value() - delta_x)
        self.verticalScrollBar().setValue(
            self.verticalScrollBar().value() - delta_y)

    def on_mid_press(self, event):
        """
        Override me
        """
        self.fit_in_scene()

    def on_wheel(self, event, has_ctrl, has_alt, has_shift):
        """
        Override me
        """
        if (has_ctrl, has_alt, has_shift) == (False, False, False):
            zoom_out = event.delta() <= 0
            self.zoom(*event.pos().toTuple(), zoom_out)

    def on_right_press(self, event):
        """
        Override me
        """
        self.setDragMode(self.ScrollHandDrag)

    def on_right_release(self, event):
        """
        Override me
        """
        self.setDragMode(self.NoDrag)

    def on_move(self, event, has_left, has_mid, has_right, this_pos, last_pos):
        """
        Override me
        """
        #
        # shift scene with right button
        if has_right:
            delta_x, delta_y = (this_pos - last_pos).toTuple()
            self.shift_view(delta_x, delta_y)



# -*- coding:utf-8 -*-


"""
"""


import sys
import argparse
#
from PySide2 import QtWidgets
#
from .bimask_app.main_window import MainWindow as BimaskMainWindow
from .dialogs import ExceptionDialog


def run_bimask_app():
    """
    """
    app = QtWidgets.QApplication(["SECV GUI"])
    mw = BimaskMainWindow()
    mw.show()
    # Wrap any exceptions into a dialog
    sys.excepthook = ExceptionDialog.excepthook
    # run app
    sys.exit(app.exec_())


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Run different GUIs.")

    parser.add_argument("--app", type=str, default="bimask",
                        help="The name of the app you want to run")
    args = parser.parse_args()
    #
    APP_NAME = args.app
    #
    if APP_NAME == "bimask":
        run_bimask_app()
    else:
        print("Unknown app name!")



# -*- coding:utf-8 -*-


"""
This module defines several reusable dialog types.
"""


import traceback
from PySide2 import QtCore, QtWidgets


# #############################################################################
# # DIALOGS
# #############################################################################
class FlexibleDialog(QtWidgets.QDialog):
    """
    Dialog class that allows for OK, Yes/No, and timeout interactions.
    To extend this dialog class, override ``setup_ui_body``, ``on_accept`` and
    ``on_reject``, store the instance and call it with ``show`` or ``exec_``.

    Note that ``setup_ui_body`` is being called IN the constructor, so any
    variables that it may need when extending the class need to be set before
    ``super().__init__`` is called.

    As it can be seen here,
    https://stackoverflow.com/questions/56449605/pyside2-qdialog-possible-bug

    implementing a Dialog in PySide2 is a little tricky. These are some things
    to consider:

    * Do not implement ``accept, reject`` directly. Rather, connect the buttons
      to ``accept, reject``, and then connect the ``accepted, rejected``
      signals to custom methods (in this case ``on_accept, on_reject``).

    * When calling the Dialog from the main window, the dialog must be
      persistently stored as a field of the main window ``i.e. self.d = ...``.
      Otherwise it will not show up. Then it can be called in modal or modeless
      way, as follows: ``XXX.connect(self.d.show), ...(self.d.exec_)``.
    """

    TIMEOUT_LBL_TXT = "Closing in {} seconds..."

    def __init__(self, accept_button_name=None, reject_button_name=None,
                 timeout_ms=None, parent=None):
        """
        :param str accept_button_name: Text to be shown in the accept button.
        :param str reject_button_name: Text to be shown in the reject button.
        :param int timeout_ms: If given, time that the dialog takes to close
          automatically, in milliseconds.
        """
        super().__init__(parent)
        # The body goes into ui_widget
        self.ui_widget = QtWidgets.QWidget()
        self.setup_ui_body(self.ui_widget)
        #
        main_layout = QtWidgets.QVBoxLayout(self)
        main_layout.addWidget(self.ui_widget)
        # Then comes the (optional) buttons section
        with_a = accept_button_name is not None
        with_r = reject_button_name is not None
        if with_a or with_r:
            button_layout = QtWidgets.QHBoxLayout()
            main_layout.addLayout(button_layout)
        #
        if with_a:
            self.accept_b = QtWidgets.QPushButton(accept_button_name)
            button_layout.addWidget(self.accept_b)
            self.accept_b.clicked.connect(self.accept)
            self.accepted.connect(self.on_accept)
            # remove autodefault, if you want default set it explicitly
            self.accept_b.setAutoDefault(False)
        #
        if with_r:
            self.reject_b = QtWidgets.QPushButton(reject_button_name)
            button_layout.addWidget(self.reject_b)
            self.reject_b.clicked.connect(self.reject)
            self.rejected.connect(self.on_reject)
            # remove autodefault, if you want default set it explicitly
            self.reject_b.setAutoDefault(False)
        # Finally the (optional) timeout message
        if timeout_ms is not None:
            assert isinstance(timeout_ms, int), \
                "Timeout miliseconds must be int or None!"
            self.timeout_lbl = QtWidgets.QLabel(
                self.TIMEOUT_LBL_TXT.format(timeout_ms / 1000))
            main_layout.addWidget(self.timeout_lbl)
        self.timeout_ms = timeout_ms

    def _set_timer(self):
        """
        """
        if self.timeout_ms is not None:
            QtCore.QTimer.singleShot(self.timeout_ms, self.reject)

    def exec_(self, *args, **kwargs):
        """
        Start the dialog in 'exclusive' way, blocking the rest of the app.
        """
        self._set_timer()
        return super().exec_(*args, **kwargs)

    def show(self, *args, **kwargs):
        """
        Start the dialog in parallel to the rest of the app.
        """
        self._set_timer()
        return super().show(*args, **kwargs)

    # OVERRIDE THESE
    def setup_ui_body(self, widget):
        """
        Populate the widget with your desired contents. The widget will
        be above the buttons.
        """
        pass

    def on_accept(self):
        """
        This method will be called if the user presses the (optional)
        accept button.
        """
        pass

    def on_reject(self):
        """
        This method will be called if the user presses the (optional)
        reject button.
        """
        pass


class InfoDialog(FlexibleDialog):
    """
    A type of dialog that shows a header and body strings.
    """

    # If true, the user can select the text with the mouse
    INTERACT_HEADER = False
    INTERACT_BODY = True

    def __init__(self, header, message, accept_button_name=None,
                 reject_button_name=None, timeout_ms=None, parent=None,
                 print_msg=True,
                 header_style="font-weight: bold; color: black"):
        """
        :param str header: This will be the dialog title
        :param str message: This will go below the title, separated by a line
        :param header_style: A CSS-like style to format the header.

        Check ``FlexibleDialog`` docstrings for more details.
        """
        self._h_txt = header
        self._msg_txt = message
        self.print_msg = print_msg
        self._h_style = header_style
        super().__init__(accept_button_name, reject_button_name, timeout_ms,
                         parent)

    def setup_ui_body(self, widget):
        """
        """
        lyt = QtWidgets.QVBoxLayout(widget)
        #
        self.header_lbl = QtWidgets.QLabel(self._h_txt)
        self.body_lbl = QtWidgets.QLabel(self._msg_txt)
        self.line = QtWidgets.QFrame()
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        #
        lyt.addWidget(self.header_lbl)
        lyt.addWidget(self.line)
        lyt.addWidget(self.body_lbl)
        if self.INTERACT_HEADER:
            self.header_lbl.setTextInteractionFlags(
                QtCore.Qt.TextSelectableByMouse)
        if self.INTERACT_BODY:
            self.body_lbl.setTextInteractionFlags(
                QtCore.Qt.TextSelectableByMouse)
        #
        self.header_lbl.setStyleSheet(self._h_style)

    def _print_if(self):
        """
        """
        if self.print_msg:
            print(self._h_txt)
            print(self._msg_txt)

    def exec_(self, *args, **kwargs):
        """
        """
        outcome = super().exec_(*args, **kwargs)
        self._print_if()
        return outcome

    def show(self, *args, **kwargs):
        """
        """
        outcome = super().show(*args, **kwargs)
        self._print_if()
        return outcome


class ExceptionDialog(InfoDialog):
    """
    This class is intended to be used at the main loop level,
    to catch any exceptions that the app may have and show them
    in a Dialog. To do that, it suffices to put the following
    line anywhere before ``app.exec_()``::

      sys.excepthook = ExceptionDialog.excepthook

    Source: https://stackoverflow.com/a/55819545/4511978
    """
    DEACTIVATE = False

    ERROR_TXT = """
ERROR!
If you think this is an app error consider reporting the following
to the developers (see Help->About):"""

    def __init__(self, error_msg, timeout_ms=None, parent=None):
        """
        Check ``InfoDialog`` docstring for details.
        """
        super().__init__(self.ERROR_TXT, error_msg, "OK",
                         "Don't show errors again",
                         header_style="font-weight: bold; color: red")

    def on_reject(self):
        """
        If the user presses on don't show errors again, the whole class
        gets deactivated, so further created instances won't pop up.
        """
        print("deactivated!")
        self.__class__.DEACTIVATE = True

    @classmethod
    def excepthook(cls, exc_type, exc_value, exc_tb):
        """
        Set this method as ``sys.excepthook = <THIS_CLASS>.excepthook``
        somewhere before ``app.exec_()`` to wrap all Python exceptions with
        this dialog.
        """
        msg = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
        print(cls.ERROR_TXT)
        print(msg)
        #
        if not cls.DEACTIVATE:
            cls(msg).exec_()



# -*- coding:utf-8 -*-


"""
This module contains the logic and widgets pertaining to the main window
of the bimask app: An app that allows displaying an image, editing a mask
on it and also displaying/editing a preannotation mask.

It can be used to efficiently annotate large images with pixel precision.
Check instructions.txt for more details.
"""


import os
from PySide2 import QtCore, QtWidgets, QtGui
import numpy as np
from PIL import Image
import json
#
from skimage.filters import apply_hysteresis_threshold
#
from .dialogs import InstructionsDialog, AboutDialog, KeymapsDialog, \
    SavedStateTracker
#
from ..masked_scene import MaskedImageScene, DisplayView
from ..base_widgets import FileList, MaskPaintForm, SaveForm
from ..utils import load_img_and_exif, unique_filename
from ..commands import DrawCommand, EraseCommand, DrawOverlappingCommand
from ..objects import PointList


# #############################################################################
# ## APPLICATION LOGIC FOR QUICK MASKING
# #############################################################################
def exp_lambda_estimator(elts):
    """
    :param elts: A collection of elements (can also be numpy).

    Given a set of elements, assumed to be exponentially distributed, returns
    the unbiased estimator for the lambda parameter of the exp distribution.
    """
    try:
        num_elts = elts.size
        elt_sum = elts.sum()
    except Exception:
        num_elts = len(elts)
        elt_sum = sum(elts)
    #
    lmbd = num_elts / elt_sum
    lmbd = lmbd - (lmbd / num_elts)
    return lmbd


def exp_threshold(keep_p_value, lmbd):
    """
    :param keep_p_value: Scalar in range ``(0, 1]``
    :param elts: collection of values, assumed to be sampled from an
      exponential distribution.
    :returns: A threshold ``t``, so that the integral for ``exp(lambda)`` from
      t to infinity equals ``keep_p_value``.

    This function assumes that the given ``elts`` have been sampled from an
    exponential distribution. Then inferes lambda, using the unbiased ML
    estimator (see https://en.wikipedia.org/wiki/Exponential_distribution)
    and returns the threshold ``t`` that fulfills ``keep_p_value`` for the
    distribution above ``t``.
    """
    assert 0 < keep_p_value <= 1, "area_p_conserve must be in (0, 1]"
    threshold = -np.log(keep_p_value) / lmbd
    return threshold


def pmap_to_mask(pmap, keep_highest_pval=0.05, discard_lowest_pval=0.5):
    """
    This method performs the following steps:

    1. Assuming that pmap values are exponentially distributed, extracts the
         unbiased lambda parameter and the cumulative distribution.
    2. Applies threshold to the given low/high p-values

    This method uses the standard connected component extraction mechanism
    in Python, i.e. ``scipy.ndimage.measurements.label`` and
    ``skimage.measure.regionprops``.

    :param pmap: A float array of shape ``h, w``.
    :param keep_highest_pval: The p-value designing the amount of top 'pmap'
      scores to be surely kept.
    :param discard_lowest_pval: The p-value designing the amount of bottom
      'pmap' scores to be surely discarded.
    :returns: The output mask
    """
    # print("PMAP TO MASK CALLED:", keep_highest_pval, discard_lowest_pval)
    assert keep_highest_pval < discard_lowest_pval, \
        "highest-P-val must be lower than lowest-P-val."

    lmbd = exp_lambda_estimator(pmap.ravel())
    t_keep_above = exp_threshold(keep_highest_pval, lmbd)
    t_discard_below = exp_threshold(discard_lowest_pval, lmbd)
    out_mask = apply_hysteresis_threshold(pmap, t_discard_below, t_keep_above)

    # sure keep: pmap>=t_keep_above   sure discard: pmap <= t_discard_below

    # t = exp_threshold(keep_below_pval, pmap.ravel())
    # mask = pmap >= t
    # # label performs connected component identification, by default
    # # using a cross-shape structuring elt.
    # labels, nlabels = label(mask)
    # # compute a set of feats for each isolated component
    # props = regionprops(labels, pmap)
    # #
    # areas = [p.area for p in props]
    # areas_threshold = exp_threshold(keep_p_value, areas)
    #
    # out_mask = np.zeros_like(mask)
    # for p in props:
    #     if p.area >= areas_threshold:
    #         coords_y, coords_x = p.coords.T
    #         out_mask[(coords_y, coords_x)] = True

    return out_mask


# #############################################################################
# ## WIDGET EXTENSIONS AND COMPOSITIONS TO ADD SPECIFIC LOGIC+LAYOUT
# #############################################################################
class FileLists(QtWidgets.QWidget):
    """
    A cluster of 3 file lists: one for images, one for masks and one for
    preannotations.
    """
    def __init__(self, parent=None, img_extensions=[".png", ".jpg", ".jpeg"],
                 mask_extensions=None, preannot_extensions=None):
        """
        If given, the extensions are case-insensitive lists in the form
        ``[".png", ".jpg"]`` that filter the files that are shown in the list
        by allowing only the given terminations.
        """
        super().__init__(parent)
        # create widgets
        self.img_list = FileList("Images\nfolder", extensions=img_extensions)
        self.mask_list = FileList("Masks\nfolder")
        self.preannot_list = FileList("Pre-annotations\nfolder")
        # add widgets to layout
        self.main_layout = QtWidgets.QHBoxLayout()
        self.main_layout.addWidget(self.img_list)
        self.main_layout.addWidget(self.mask_list)
        self.main_layout.addWidget(self.preannot_list)
        self.setLayout(self.main_layout)


class IntegratedSaveForm(SaveForm):
    """
    A ``SaveForm`` that implements this app's logic, namely, it features 2
    masks, one for annot and one for preannot, and saves them as B&W png.
    """
    def __init__(self, main_window, default_path=None,
                 save_dialog_timeout_ms=1000):
        """
        :param main_window: A reference to the ``BimaskMainWindow``
        :param str default_path: If non given, 'home' is picked.
        :param save_dialog_timeout: When successfully saving, a dialog
          will pop up, and disappear after this many miliseconds.
        """
        super().__init__(None, default_path)
        self.main_window = main_window
        self.add_checkbox("preannot.", initial_val=False,
                          initial_txt="_preannot.png")
        self.add_checkbox("annot.", initial_txt="_annot.png")
        self.add_checkbox("points", initial_txt="_points.json")
        # This reference is needed otherwise dialogs get garbage collected?
        self.dialog = None
        self.dialog_ms = save_dialog_timeout_ms

    def save_masks(self, states, suffixes, overwrite):
        """
        Overriden method that we don't call directly. See ``SaveForm`` for
        interface details.
        """
        save_preannot, save_annot, save_points = states
        suff_preannot, suff_annot, suff_points = suffixes
        img_name = self.main_window.current_img_basename
        #
        a_pmi = self.main_window.graphics_view.annot_pmi
        pa_pmi = self.main_window.graphics_view.preannot_pmi
        #
        scene = self.main_window.graphics_view.scene()
        saved = {}
        if save_preannot and pa_pmi is not None:
            pa_path = os.path.join(self.save_path, img_name + suff_preannot)
            if not overwrite:
                pa_path = unique_filename(pa_path)
            pa_msk_arr = scene.mask_as_bool_arr(pa_pmi)
            self.save_bool_arr_as_img(pa_msk_arr, pa_path, overwrite)
            saved["preannotation mask"] = pa_path
        if save_annot and a_pmi is not None:
            a_path = os.path.join(self.save_path, img_name + suff_annot)
            if not overwrite:
                a_path = unique_filename(a_path)
            msk_arr = scene.mask_as_bool_arr(a_pmi)
            self.save_bool_arr_as_img(msk_arr, a_path, overwrite)
            saved["annotation mask"] = a_path
        if save_points and scene.objects:
            state_dict = {k.__name__: [elt.state() for elt in v if elt.state()]
                          for k, v in scene.objects.items()}
            p_path = os.path.join(self.save_path, img_name + suff_points)
            if not overwrite:
                p_path = unique_filename(p_path)
            with open(p_path, "w") as f:
                # f.write(str(state_dict))
                json.dump(state_dict, f)
            saved["point lists"] = p_path
        #
        if saved:
            self.main_window.graphics_view.saved_state_tracker.save(
                saved, self.dialog_ms)

    def save_bool_arr_as_img(self, arr, outpath, overwrite_existing=False):
        """
        Output: RGB PNG image where false is black (0, 0, 0) and true is white
        (255, 255, 255).
        """
        if not overwrite_existing:
            outpath = unique_filename(outpath)
        img = Image.fromarray(arr)
        img.save(outpath)


class IntegratedDisplayView(DisplayView):
    """
    This class implements the main component of the main window: it features a
    view of the image and the masks, together with a set of operations that can
    be done on them (painting, updating...), and the callback mechanisms to
    trigger those operations.
    """
    def __init__(self, main_window, scale_percent=15):
        """
        :param scale_percent: Each zoom in/out operation will scale the view
          by this much (in percent).
        """
        super().__init__(scene=None, parent=None, scale_percent=scale_percent)
        self._scene = MaskedImageScene()
        self.main_window = main_window
        self.shape = None
        self.setScene(self._scene)
        #
        self._preannot_pmap = None
        self.preannot_pmi = None
        self.annot_pmi = None
        #
        #
        self._current_clickdrag_action = None
        #
        self.saved_state_tracker = None

    # MEMORY ACTIONS
    def new_image(self, img_path, initial_mask_color=(219, 54, 148, 150),
                  initial_preannot_color=(102, 214, 123, 100)):
        """
        If successful, removes all elements from the scene and the undo stack,
        and loads a fresh image and masks. If there are unsaved changes, a
        dialog asking for confirmation will pop up.

        :returns: True if the action completed successfully, False if the user
          decides to abort.
        """
        if self.saved_state_tracker is not None:
            is_delete_ok = self.saved_state_tracker.delete()
            if not is_delete_ok:
                # If user didn't want to delete unsaved changes
                return False
        # Go on with the update
        img_arr = load_img_and_exif(img_path)[0]
        self.shape = img_arr.shape
        self._scene.update_image(img_arr)
        dummy_preannot = np.zeros(img_arr.shape[:2], dtype=np.bool)
        dummy_mask = np.zeros_like(dummy_preannot)
        self.preannot_pmi = self._scene.add_mask(
            dummy_preannot, initial_preannot_color)
        self.annot_pmi = self._scene.add_mask(
            dummy_mask, initial_mask_color)
        self.fit_in_scene()
        #
        self.main_window.undo_stack.clear()
        #
        self.saved_state_tracker = SavedStateTracker()
        return True

    def preannot_from_path(self, preannot_path, rgba, keep_p_value=0.05,
                           discard_p_value=0.5, normalize=False):
        """
        This method is prototype-ish: It loads an ``.npz`` file with and
        'entropy' field, expected to have a numpy float matrix with same
        shape as the image.
        """
        assert self.scene().img_pmi is not None, \
            "You need to load an image first!"
        self._preannot_pmap = np.load(preannot_path)["entropy"]
        if normalize:
            try:
                self._preannot_pmap /= self._preannot_pmap.max()
            except ZeroDivisionError:
                pass
        m = pmap_to_mask(self._preannot_pmap, keep_p_value, discard_p_value)
        self.preannot_pmi = self.scene().replace_mask_pmi(
            self.preannot_pmi, m)
        #
        self.saved_state_tracker.edit()

    def mask_from_path(self, mask_path, rgba):
        """
        :param mask_path: Path to an image containing a binary mask, where
          zero pixels are considered false and non-zero true.
        :param rgba: Color of the loaded mask

        Loads a binary mask into the scene as an RGBA-colored mask.
        """
        assert self.scene().img_pmi is not None, \
            "You need to load an image first!"
        arr = load_img_and_exif(mask_path)[0]
        if len(arr.shape) == 2:
            mask = arr > 0
        elif len(arr.shape) == 3:
            mask = arr.any(axis=-1)
        else:
            raise RuntimeError("Mask must be rank 2 or 3!")
        self.annot_pmi = self.scene().replace_mask_pmi(
            self.annot_pmi, mask)
        #
        self.saved_state_tracker.edit()

    # MASK SINGLE-SHOT ACTIONS
    def change_preannot_pval(self, keep_p_value, discard_p_value=0.5):
        """
        Updates the preannot->mask threshold.
        """
        if self._preannot_pmap is not None:
            new_m = pmap_to_mask(self._preannot_pmap, keep_p_value,
                                 discard_p_value)
            self.preannot_pmi = self.scene().replace_mask_pmi(
                self.preannot_pmi, new_m)
        #
        self.saved_state_tracker.edit()

    def change_preannot_rgba(self, rgba):
        """
        Updates the preannot mask color.
        """
        if self.preannot_pmi is not None:
            m = self.scene().mask_as_bool_arr(self.preannot_pmi)
            self.preannot_pmi = self.scene().replace_mask_pmi(
                self.preannot_pmi, m, rgba)

    def change_annot_rgba(self, rgba):
        """
        Updates the annot mask color.

        """
        if self.annot_pmi is not None:
            m = self.scene().mask_as_bool_arr(self.annot_pmi)
            self.annot_pmi = self.scene().replace_mask_pmi(
                self.annot_pmi, m, rgba)

    # MASK COMPOSITE ACTIONS
    def _finish_clickdrag_action(self):
        """
        finishes any click+drag action that may be active (does nothing if
        none active).
        """
        cmd = self._current_clickdrag_action
        if cmd is not None:
            cmd.finish(self.main_window.undo_stack)
            self._current_clickdrag_action = None

    def _perform_composite_action(self, action_class, action_args,
                                  construction_args):
        """
        This function is the recommended way to perform a composite
        action for the following reasons:

        1. If ``action_class`` is already running, it simply continues it.
        2. If a different composite action was running, it closes it and starts
            this one.
        3. If no composite action was running, starts this one

        And finally performs the action.

        :param construction_args: If this action needs to be started, it will
          be called via ``cmd = action_class(*construction_args)``
        :param action_args: The command will be called via ``cmd(action_args)``

        Usage example::
          x, y = current_action_position...
          pmi = ...
          brush_size = ...
          rgba = self.scene().mask_pmis[pmi]
          self._perform_composite_action(DrawCommand, [x, y],
                                         [pmi, rgba, brush_size])
        """
        cmd = self._current_clickdrag_action
        # if changed to this action without releasing the prior one, release it
        action_changed = action_class is not cmd.__class__
        cmd_finished = cmd is not None and cmd.finished
        if action_changed:
            self._finish_clickdrag_action()  # sets current action to None
            cmd = self._current_clickdrag_action
        # if no open action of this class, create
        if cmd is None or cmd_finished:
            cmd = action_class(*construction_args)
            self._current_clickdrag_action = cmd
        cmd.action(*action_args)

    def clickdrag_action(self, x, y):
        """
        Paint to the currently selected mask, with the currently selected
        brush type, at the given position.
        The given ``x, y`` position is in 'scene coordinates', i.e. the
        position from a mouse event has to be translated as follows::

          xpos, ypos = self.mapToScene(event.pos()).toTuple()
          self.clickdrag_action(xpos, ypos)
        """
        # retrieve pmi info
        # expected idx: 0 for preannot, 1 for annot
        idx_map = {0: self.preannot_pmi, 1: self.annot_pmi}
        mask_idx = self.main_window.paint_form.current_button_idx
        pmi = idx_map[mask_idx]
        # paint only if this pmi
        if pmi is None:
            return
        # retrieve brush info
        p_txt, e_txt, mp_txt = [self.main_window.PAINTER_TXT,
                                self.main_window.ERASER_TXT,
                                self.main_window.MASKED_PAINTER_TXT]
        brush_type = self.main_window.paint_form.current_brush_type
        brush_size = self.main_window.paint_form.current_brush_size
        # if no open action exists, create:
        did_something = False
        if brush_type == p_txt:
            rgba = self.scene().mask_pmis[pmi]
            self._perform_composite_action(DrawCommand, [x, y],
                                           [pmi, rgba, brush_size])
            did_something = True
        elif brush_type == e_txt:
            self._perform_composite_action(EraseCommand, [x, y],
                                           [pmi, brush_size])
            did_something = True
        elif brush_type == mp_txt:
            rgba = self.scene().mask_pmis[pmi]
            ref_pmi = self.preannot_pmi  # preannot is always the ref
            self._perform_composite_action(DrawOverlappingCommand, [x, y],
                                           [pmi, ref_pmi, rgba, brush_size])
            did_something = True
        #
        if did_something:
            self.saved_state_tracker.edit()

    def add_point(self, x, y, close_after=False):
        """
        """
        if self.scene().img_pmi is None:
            return
        brush_size = self.main_window.paint_form.current_brush_size
        self.scene().object_action(
            PointList, [x, y, self.main_window.undo_stack],
            [self.scene(), brush_size, (0, 0, 0, 100), (0, 0, 0, 255),
             True])  # draw lines
        #
        if close_after:
            self.scene().close_current_object_action(
                self.main_window.undo_stack)

    # EVENT HANDLING
    def on_left_press(self, event):
        """
        """
        xpos, ypos = self.mapToScene(event.pos()).toTuple()
        brush_type = self.main_window.paint_form.current_brush_type
        if brush_type == self.main_window.POINT_LIST_TXT:
            mods = event.modifiers()
            has_ctrl = bool(mods & QtCore.Qt.ControlModifier)
            self.add_point(xpos, ypos, close_after=has_ctrl)
        else:
            self.clickdrag_action(xpos, ypos)

    def on_left_release(self, event):
        """
        If there is an open macro command, closes it and adds it to the undo
        stack
        """
        self._finish_clickdrag_action()

    def on_move(self, event, has_left, has_mid, has_right, this_pos, last_pos):
        """
        Callback implementation, calls ``clickdrag_action`` if moving while
        pressing left.
        """
        super().on_move(event, has_left, has_mid, has_right, this_pos,
                        last_pos)
        #
        if has_left:
            xpos, ypos = self.mapToScene(event.pos()).toTuple()
            self.clickdrag_action(xpos, ypos)


class CrackAnnotPaintForm(MaskPaintForm):
    """
    A ``MaskPaintForm`` that holds a reference to the app's main window and
    connects its callbacks with the main window's corresponding components.
    """
    def __init__(self, main_window, brushes, max_brush_size=100, parent=None,
                 thresh_min=0, thresh_max=1, thresh_num_steps=100):
        """
        :param main_window: A reference to the bimask app main window instance.
        :param brushes: A list of brush names to be featured in the form.
        """
        super().__init__(brushes, max_brush_size, parent, thresh_min,
                         thresh_max, thresh_num_steps, min_alpha=1)
        self.main_window = main_window
        self.current_brush_type = self.brush_names[
            self.brush_combo_box.currentIndex()]
        self.current_brush_size = self.brush_size_slider.value()
        self.current_button_idx = None  # activate when calling addItem

    def button_pressed(self, but):
        """
        Setter
        """
        self.current_button_idx = self._buttons.index(but)

    def threshold_slider_changed(self, idx, val):
        """
        :param int idx: The mask index. 0 is the index of the preannotation,
          1 for annotation.
        :param val: The new p-value.

        Update preannotation mask with new p-value by calling the
        ``change_preannot_val`` method of the view. Only works if idx is 0.
        """
        if idx == 0:
            self.main_window.graphics_view.change_preannot_pval(
                val, self.main_window.DISCARD_P_VALUE)

    def rgba_box_changed(self, idx, r, g, b, a):
        """
        Update corresponding mask with new RGBA color.
        """
        # NOTE: THIS ASSUMES THAT BOX 0 IS ANNOT AND BOX 1 IS PREANNOT!
        assert idx in {0, 1}, "This GUI wasn't prepared for more than 2 masks"
        # self.main_window.graphics_view.scene().change_mask_color(
        #     idx, r, g, b, a)
        view = self.main_window.graphics_view
        # idx_map = {0: view.preannot_pmi, 1: view.annot_pmi}
        # pmi = idx_map[idx]
        if idx == 0:
            view.change_preannot_rgba((r, g, b, a))
            self.main_window.preannot_color = (r, g, b, a)
        elif idx == 1:
            view.change_annot_rgba((r, g, b, a))
            self.main_window.mask_color = (r, g, b, a)

    def brush_type_changed(self, idx):
        """
        Setter
        """
        self.current_brush_type = self.brush_names[idx]

    def brush_size_changed(self, sz):
        """
        Setter
        """
        self.current_brush_size = sz


# #############################################################################
# ## MAIN WINDOW
# #############################################################################
class MainWindow(QtWidgets.QMainWindow):
    """
    This is the central widget for the bimask application. It is a composition
    of all the used elements, together with the logic that binds them.
    """

    # These variables handle the preannotation thresholding. Check pmap_to_mask
    DISCARD_P_VALUE = 0.5  # Number in range (thresh_slider_max, 1]
    THRESH_MIN = 0.0000001
    THRESH_MAX = 0.000000001
    THRESH_NUM_STEPS = 400
    #
    PAINTER_TXT = "Painter"
    ERASER_TXT = "Eraser"
    MASKED_PAINTER_TXT = "Masked painter"
    POINT_LIST_TXT = "Points"

    def __init__(self, parent=None, initial_mask_color=(255, 54, 76, 150),
                 initial_preannot_color=(102, 214, 123, 100),
                 max_brush_size=200):
        """
        """
        super().__init__(parent)
        self.graphics_view = IntegratedDisplayView(self)
        #
        self.mask_color = initial_mask_color
        self.preannot_color = initial_preannot_color
        #
        self.current_img_basename = None
        #
        self.instructions_dialog = InstructionsDialog()
        self.about_dialog = AboutDialog()
        self.keymaps_dialog = KeymapsDialog(
            {k: v.toString() for k, v in self.keymaps().items()})
        # define controller widgets
        self.file_lists = FileLists()
        self.paint_form = CrackAnnotPaintForm(
            self, [self.PAINTER_TXT, self.ERASER_TXT, self.MASKED_PAINTER_TXT,
                   self.POINT_LIST_TXT],
            max_brush_size, thresh_min=self.THRESH_MIN,
            thresh_max=self.THRESH_MAX, thresh_num_steps=self.THRESH_NUM_STEPS)
        self.save_form = IntegratedSaveForm(self, default_path=None)

        self.paint_form.add_item("preannot.", self.preannot_color,
                                 slider_visible=True, activate=False)
        self.paint_form.add_item("annot.", self.mask_color,
                                 slider_visible=False, activate=True)
        # create controller layout
        controller_layout = QtWidgets.QVBoxLayout()
        controller_layout.addWidget(self.paint_form)
        controller_layout.addWidget(self.save_form)
        controller_widget = QtWidgets.QWidget()
        controller_widget.setLayout(controller_layout)
        self.controller_splitter = QtWidgets.QSplitter()
        self.controller_splitter.setOrientation(QtCore.Qt.Vertical)
        self.controller_splitter.addWidget(self.file_lists)
        self.controller_splitter.addWidget(controller_widget)
        # create main layout, add controller and graphics:
        self.main_splitter = QtWidgets.QSplitter()
        self.main_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.main_splitter.addWidget(self.controller_splitter)
        self.main_splitter.addWidget(self.graphics_view)
        # fine-tune main layout: sizes and such
        self.controller_splitter.setMinimumWidth(10)
        left_width = self.controller_splitter.width()
        right_width = self.graphics_view.width()
        self.main_splitter.setSizes([left_width, right_width * 2])
        self.setCentralWidget(self.main_splitter)
        # add connections
        self.file_lists.img_list.file_list.itemDoubleClicked.connect(
            lambda elt: self._handle_img_selection(elt.text()))
        self.file_lists.mask_list.file_list.itemDoubleClicked.connect(
            lambda elt: self._handle_mask_selection(elt.text()))
        self.file_lists.preannot_list.file_list.itemDoubleClicked.connect(
            lambda elt: self._handle_preannot_selection(elt.text()))
        #
        self._setup_undo()
        self._setup_menu_bar()
        self._add_keymaps()

    def _setup_undo(self):
        """
        Set up undo stack and undo view
        """
        self.undo_stack = QtWidgets.QUndoStack(self)
        self.undo_view = QtWidgets.QUndoView(self.undo_stack)
        self.undo_view.setWindowTitle("Undo View")
        self.undo_view.setAttribute(QtCore.Qt.WA_QuitOnClose, False)

    def _setup_menu_bar(self):
        """
        Set up menu bar: create actions and connect them to methods.
        """
        # edit menu
        edit_menu = self.menuBar().addMenu("Edit")
        self.undo_action = edit_menu.addAction("Undo")
        self.undo_action.triggered.connect(self.undo_stack.undo)
        self.redo_action = edit_menu.addAction("Redo")
        self.redo_action.triggered.connect(self.undo_stack.redo)
        edit_menu.addSeparator()
        self.view_undo_action = edit_menu.addAction("View undo stack")
        self.view_undo_action.triggered.connect(self.undo_view.show)
        # help menu
        help_menu = self.menuBar().addMenu("Help")
        self.keyboard_shortcuts = help_menu.addAction("Keyboard shortcuts")
        self.keyboard_shortcuts.triggered.connect(self.keymaps_dialog.show)
        self.instructions = help_menu.addAction("Instructions")
        self.instructions.triggered.connect(self.instructions_dialog.show)
        self.about = help_menu.addAction("About")
        self.about.triggered.connect(self.about_dialog.show)

    def keymaps(self):
        """
        :returns: A dictionary in the form ``name: QtGui.QKeySequence``,
          where the

        Define this GUI's specific key mappings. Note that this method can
        be overriden to return a different mapping, but the ``name``s have
        to remain identical, in order to be recognized by ``_add_keymaps``.
        """
        d = {
            "Undo": QtGui.QKeySequence("Ctrl+Z"),
            "Redo": QtGui.QKeySequence("Ctrl+Y"),
            "View undo list": QtGui.QKeySequence("Alt+Z"),
            #
            "Load image path": QtGui.QKeySequence("Ctrl+I"),
            "Load mask path": QtGui.QKeySequence("Ctrl+M"),
            "Load preannotation path": QtGui.QKeySequence("Ctrl+P"),
            #
            "Save mask path": QtGui.QKeySequence("Alt+S"),
            "Save mask(s)": QtGui.QKeySequence("Ctrl+S"),
            #
            "Set painter": QtGui.QKeySequence("a"),
            "Set eraser": QtGui.QKeySequence("e"),
            "Set masked painter": QtGui.QKeySequence("m"),
            #
            "Next image": QtGui.QKeySequence("Space"),
            "Previous image": QtGui.QKeySequence("Ctrl+Space")
        }
        return d

    def _add_keymaps(self):
        """
        This function is closeley connected to ``keymaps``. There, the
        shortcuts are defined, here, they are applied.
        """
        km = self.keymaps()
        # add menu shortcuts
        self.undo_action.setShortcut(km["Undo"])
        self.redo_action.setShortcut(km["Redo"])
        self.view_undo_action.setShortcut(km["View undo list"])
        # add widget shortcuts
        #
        self.file_lists.img_list.file_button.setShortcut(km["Load image path"])
        self.file_lists.mask_list.file_button.setShortcut(km["Load mask path"])
        self.file_lists.preannot_list.file_button.setShortcut(
            km["Load preannotation path"])
        #
        self.save_form.file_dialog_button.setShortcut(km["Save mask path"])
        self.save_form.save_button.setShortcut(km["Save mask(s)"])
        # Paint region (wheel event has the brush size)
        QtWidgets.QShortcut(  # combobox shortcuts are a little more complex
            km["Set painter"], self.paint_form.brush_combo_box,
            lambda: self.paint_form.brush_combo_box.setCurrentText(
                self.PAINTER_TXT))
        QtWidgets.QShortcut(
            km["Set eraser"], self.paint_form.brush_combo_box,
            lambda: self.paint_form.brush_combo_box.setCurrentText(
                self.ERASER_TXT))
        QtWidgets.QShortcut(
            km["Set masked painter"], self.paint_form.brush_combo_box,
            lambda: self.paint_form.brush_combo_box.setCurrentText(
                self.MASKED_PAINTER_TXT))
        #
        QtWidgets.QShortcut(
            km["Next image"], self, lambda: self._switch_img(1))
        QtWidgets.QShortcut(
            km["Previous image"], self, lambda: self._switch_img(-1))

    def _switch_img(self, step=1):
        """
        An alternative way of double clicking on an image list is to call
        this method, which will switch to the image located at
        ``curent_img + step`` in the list.
        """
        curr_idx = self.file_lists.img_list.file_list.currentRow()
        nxt_item = self.file_lists.img_list.file_list.item(curr_idx + step)
        if nxt_item is not None:
            success = self._handle_img_selection(nxt_item.text())
            if success:
                self.file_lists.img_list.file_list.setCurrentItem(nxt_item)

    def _handle_img_selection(self, basename):
        """
        This protected method is triggered when double clicking on an
        image list item, or called by ``_switch_img``.
        """
        abspath = os.path.join(self.file_lists.img_list.dirpath, basename)
        success = self.graphics_view.new_image(abspath, self.mask_color,
                                               self.preannot_color)
        if success:
            self.current_img_basename = basename
        return success

    def _handle_mask_selection(self, basename):
        """
        This protected method is triggered when double clicking on an
        annotation list item.
        """
        abspath = os.path.join(self.file_lists.mask_list.dirpath, basename)
        self.graphics_view.mask_from_path(abspath, self.mask_color)

    def _handle_preannot_selection(self, basename):
        """
        This protected method is triggered when double clicking on a
        preannotation list item.
        """
        abspath = os.path.join(self.file_lists.preannot_list.dirpath, basename)
        pval = self.paint_form.slider_to_p_val(
            self.paint_form._sliders[-1].value())
        self.graphics_view.preannot_from_path(
            abspath, self.preannot_color, keep_p_value=pval,
            discard_p_value=self.DISCARD_P_VALUE)

    def wheelEvent(self, event):
        """
        The ``DisplayView`` has zoom functionality associated to the wheel.
        Here we associate 'brush size change' functionality when the wheel
        is rolled while pressing Control.
        """
        mods = event.modifiers()
        has_ctrl = bool(mods & QtCore.Qt.ControlModifier)
        has_alt = bool(mods & QtCore.Qt.AltModifier)
        has_shift = bool(mods & QtCore.Qt.ShiftModifier)
        if (has_ctrl, has_alt, has_shift) == (True, False, False):
            current = self.paint_form.brush_size_slider.value()
            delta = 1 if event.delta() >= 0 else - 1
            self.paint_form.brush_size_slider.setValue(current + delta)



# -*- coding:utf-8 -*-


"""
This module contains all the 'undoable' actions. They must implement a way
to undo and redo them.

Composite commands deserve a special mention: they are trains of actions
that only track, store and report the initial and final state. They are
particularly useful when performing interactive editings on big datastructures
like pixmaps, to prevent memory bloating.
"""


import numpy as np
from PySide2 import QtCore, QtWidgets, QtGui
from .utils import pixmap_to_arr


# #############################################################################
# ## SINGLE-SHOT COMMANDS
# #############################################################################
class UndoableLambda(QtWidgets.QUndoCommand):
    """
    This kind of functor can be used to create them on the spot and send them
    to the UndoStack. Useful to split down a composite action in arbitrary
    undo-able subactions. Usage example::

      # in action ... do something ...
      cmd = UndoableLambda("My partial action",
                           lambda: print("undo"), lambda: print("redo"))
      undo_stack.push(cmd)
    """
    def __init__(self, command_name, undo_fn, redo_fn, parent=None):
        """
        """
        super().__init__(command_name, parent)
        self.undo_fn = undo_fn
        self.redo_fn = redo_fn

    def undo(self):
        """
        """
        self.undo_fn()

    def redo(self):
        """
        """
        self.redo_fn()

# #############################################################################
# ## SINGLE-SHOT COMMANDS
# #############################################################################


# #############################################################################
# ## COMPOSITE COMMANDS
# #############################################################################
class CompositeCommand(QtWidgets.QUndoCommand):
    """
    In some cases like painting a stroke into a pixmap, it doesn't make sense
    to store every single update: rather, the prior and finished states only.
    This class provides a structure for such cases:

    1. Instantiate the command with the parameters that belong to the whole
        composite action.
    2. Call ``action`` for every desired update of the finished state
    3. Call ``finish`` to crystalize the final state. No further ``action`` s
        will be allowed, and (optionally) the action will be added to a Qt
        ``UndoStack``.

    The following is required to extend the class:
    1. Define a ``COMMAND_NAME``
    2. Extend the ``__init__``, ``action``, ``undo`` and ``redo`` methods.
    """
    # override this when extending!
    COMMAND_NAME = NotImplemented

    def __init__(self, parent=None):
        """
        """
        super().__init__(self.COMMAND_NAME, parent)
        self.finished = False

    def action(self):
        """
        Extend me!
        """
        assert not self.finished, \
            "This CompositeCommand has already been finished!"

    def finish(self, undo_stack=None):
        """
        :param undo_stack: If given, this command will be added to the stack,
          wich then allows to undo/redo.

        Call this function once you are done with the ``action`` s. Once finish
        is called, no more ``action`` s are possible, so that the undo/redo
        actions stay frozen.
        """
        self.finished = True
        if undo_stack is not None:
            undo_stack.push(self)


class DrawCommand(CompositeCommand):
    """
    A composite command to draw a stroke of circles into a ``PixmapIten``.
    """
    COMMAND_NAME = "Draw"

    def __init__(self, pmi, rgba, diameter,
                 comp_mode=QtGui.QPainter.CompositionMode_Source, parent=None):
        """
        :param pmi: A ``PixmapItem``, where this command will apply.
        :param rgba: A tuple ``(r, g, b, a)`` of ``0-255`` values.
        :param diameter: In pixels, diameter of the circle to be drawn.
        :param comp_mode: The "Source" mode ensures that the alphas don't
          get added.
        """
        super().__init__(parent)
        # Caution: memory intensive? (<100 commands on 4k*6k seems ok)
        self.original_pixmap = pmi.pixmap()
        self.final_pixmap = pmi.pixmap()
        #
        self.pmi = pmi
        self.rgba = rgba
        self.diameter = diameter
        self.comp_mode = comp_mode
        #
        color = QtGui.QColor(*rgba)
        self.brush = QtGui.QBrush(color, bs=QtCore.Qt.SolidPattern)
        self.pen = QtGui.QPen(color)

    def action(self, x_pos, y_pos):
        """
        Once the object has been constructed **and ``finish()`` hasn't been
        called yet**, Call this function to paint a circle at given position.
        Check constructor for further variables.
        """
        super().action()
        # get painter and set it up
        painter = QtGui.QPainter(self.final_pixmap)
        painter.setCompositionMode(self.comp_mode)
        painter.setBrush(self.brush)
        painter.setPen(self.pen)
        #
        diameter = self.diameter
        radius = diameter // 2
        xywh = (x_pos - radius, y_pos - radius, diameter, diameter)
        painter.drawEllipse(QtCore.QRect(*xywh))
        painter.end()
        self.pmi.setPixmap(self.final_pixmap)

    def redo(self):
        """
        This function implements the interface for the UndoStack. Don't call
        this directly.
        """
        self.pmi.setPixmap(self.final_pixmap)

    def undo(self):
        """
        This function implements the interface for the UndoStack. Don't call
        this directly.
        """
        self.pmi.setPixmap(self.original_pixmap)

    def finish(self, undo_stack=None):
        """
        Usually we don't override ``finish``, but since pixmaps are so big,
        we don't want to store the command if original and final are equal.
        """
        self.finished = True
        if undo_stack is not None:
            original_arr = pixmap_to_arr(self.original_pixmap,
                                         QtGui.QImage.Format_RGBA8888)
            final_arr = pixmap_to_arr(self.final_pixmap,
                                      QtGui.QImage.Format_RGBA8888)
            # Add to stack only if there is any difference
            if not np.array_equal(original_arr, final_arr):
                undo_stack.push(self)


class EraseCommand(DrawCommand):
    """
    A composite command to erase a stroke of circles into a ``PixmapIten``.
    See ``DrawCommand`` docstrings for more info.
    """
    COMMAND_NAME = "Erase"

    def __init__(self, pmi, diameter,
                 comp_mode=QtGui.QPainter.CompositionMode_Source, parent=None):
        """
        See ``DrawCommand`` docstrings for more info.
        """
        super().__init__(pmi, (0, 0, 0, 0), diameter, comp_mode)


class DrawOverlappingCommand(DrawCommand):
    """
    Like ``DrawCommand``, but accepts 2 PixmapItems instead of one, so that
    the drawing onto the first is only allowed if the same pixel is active in
    the second.
    """
    COMMAND_NAME = "Draw Overlapping"

    def __init__(self, pmi, ref_pmi, rgba, diameter,
                 comp_mode=QtGui.QPainter.CompositionMode_Source, parent=None):
        """
        :param ref_pmi: This ``PixmapItem`` should be of same shape as ``pmi``.
        See ``DrawCommand`` docstrings for more info.
        """
        super().__init__(pmi, rgba, diameter, comp_mode)
        self._reference_pmi = ref_pmi

    def action(self, x_pos, y_pos):
        """
        Paint a circle on ``pmi`` at given position, masked by ``ref_pmi``.
        """
        # get painter and set it up
        painter = QtGui.QPainter(self.final_pixmap)
        painter.setCompositionMode(self.comp_mode)
        painter.setBrush(self.brush)
        painter.setPen(self.pen)
        #
        diameter = self.diameter
        radius = diameter // 2
        xywh = (x_pos - radius, y_pos - radius, diameter, diameter)
        #
        # mask out the relevant region from the other mask: for that find the
        # zeros, and mask out anything that IS a zero
        other_pm = self._reference_pmi.pixmap().copy(QtCore.QRect(*xywh))
        other_mask = other_pm.createMaskFromColor(
            QtGui.QColor(0, 0, 0, a=0), mode=QtCore.Qt.MaskInColor)
        other_region = QtGui.QRegion(other_mask)
        # translate region to global coords and apply mask to painter
        other_region.translate(x_pos - radius, y_pos - radius)
        painter.setClipRegion(other_region)
        # draw and refresh
        painter.drawEllipse(QtCore.QRect(*xywh))
        painter.end()
        self.pmi.setPixmap(self.final_pixmap)
